{ File   : DATRead.dpr
  Author : Deniz Oezmen
  Created: 2006-10-20
  Changed: 2010-03-22

  Dreamfall localization.dat converter.
}
program DATRead;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TLangHeader  = record
                   AbbrevLen: Longint;
                   Abbrev   : string;
                   NameLen  : Longint;
                   Name     : string;
                   NumLines,
                   Size     : Longint;
                 end;

  TLangLine    = packed record
                   ID,
                   Offset: Longint;
                 end;

  TLangLines = array of TLangLine;

var
  DATFile: file;

function OpenDATFile(const DATName: string; var DATFile: file): Boolean;
begin
  AssignFile(DATFile, DATName);
  {$I-}
  Reset(DATFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

procedure ReadDAT(var DATFile: file);
var
  i,
  BaseOfs,
  CurPos     : Longint;
  c          : Char;
  s          : string;
  LangHeader : TLangHeader;
  LangLines  : TLangLines;
  LangNums   : array of Longint;
  LangStrings: array of string;
begin
  while not Eof(DATFile) do
  begin
    with LangHeader do
    begin
      BlockRead(DATFile, AbbrevLen, 4);
      SetLength(Abbrev, AbbrevLen);
      BlockRead(DATFile, Abbrev[1], AbbrevLen);
      BlockRead(DATFile, NameLen, 4);
      SetLength(Name, NameLen);
      BlockRead(DATFile, Name[1], NameLen);

      if (AbbrevLen or NameLen) = 0 then
        Continue;

      BlockRead(DATFile, NumLines, 4);
      BlockRead(DATFile, Size, 4)
    end;

    WriteLn(Format('abbrev=%s', [LangHeader.Abbrev]));
    WriteLn(Format('name=%s', [LangHeader.Name]));
    WriteLn(Format('numlines=%d', [LangHeader.NumLines]));

    SetLength(LangLines, LangHeader.NumLines);
    BlockRead(DATFile, LangLines[0],
      SizeOf(TLangLine) * LangHeader.NumLines);
    SetLength(LangNums, LangHeader.NumLines);

    BaseOfs := FilePos(DATFile);

    SetLength(LangStrings, 0);

    while FilePos(DATFile) < BaseOfs + LangHeader.Size do
    begin
      CurPos := FilePos(DATFile);
      for i := 0 to LangHeader.NumLines - 1 do
        if BaseOfs + LangLines[i].Offset = CurPos then
          LangNums[i] := High(LangStrings) + 1;

      s := '';
      c := #255;
      while c <> #000 do
      begin
        BlockRead(DATFile, c, 1);
        s := s + c
      end;
      SetLength(s, Length(s) - 1);

      s := AnsiReplaceStr(s, #013#010, '\n');
      s := AnsiReplaceStr(s, #013, '\d');
      s := AnsiReplaceStr(s, #010, '\a');

      SetLength(LangStrings, Length(LangStrings) + 1);
      LangStrings[High(LangStrings)] := s
    end;

    WriteLn(Format('numstrings=%d', [Length(LangStrings)]));

    for i := 0 to High(LangLines) do
      WriteLn(Format('%d=%d', [LangLines[i].ID, LangNums[i]]));

    for i := 0 to High(LangStrings) do
      WriteLn(LangStrings[i]);

    Seek(DATFile, BaseOfs + LangHeader.Size)
  end
end;

begin
  if ParamCount <> 1 then
  begin
    WriteLn('Incorrect number of parameters');
    Exit
  end;

  if not OpenDATFile(ParamStr(1), DATFile) then
  begin
    WriteLn('Could not open input file.');
    Exit
  end;

  ReadDAT(DATFile);

  CloseFile(DATFile)
end.

