{ File   : MakePAK.dpr
  Author : Deniz Oezmen
  Created: 2007-03-29
  Changed: 2010-03-22

  Dreamfall PAK creator.
}
program MakePAK;

{$APPTYPE CONSOLE}

uses
  Windows, SysUtils, QSort;

const
  NameTable: array[1..43] of Char =
    'abcdefghijklmnopqrstuvwxyz/'#010#013'-_''.0123456789';
  NameTableSet: set of Char =
    ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j','k', 'l', 'm', 'n', 'o',
     'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '/', #010, #013,
     '-', '_', '''',  '.', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9'];
  // as generated by dreamfall.exe
  NameTableInv: array[0..255] of Byte =
    (32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 28, 32, 32, 29, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 30, 33, 27,
     34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 32, 32, 32, 32, 32, 32,
     32,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
     16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 32, 27, 32, 32, 31,
     32,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
     16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32,
     32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32, 32);

type
  TFileInfo  = record
                 RealName,
                 FSName  : string[255];
                 Size,
                 NodeIdx : Longint
               end;

  TFileInfos = array of TFileInfo;

  TPAKHeader  = packed record
                  ID          : array[0..11] of Char;
                  NumHashNodes,
                  NumNameIdxs,
                  NumNames    : Longint
                end;

  THashNode  = packed record
                 Offset,
                 Size,
                 BaseNodeIdx,
                 PathLen,
                 NameIdx    : Longint
               end;

  THashNodes = array of THashNode;

  TNames     = array of Byte;

  TNameIdxs  = array of Longint;

  TAllocated = array of ByteBool;

function OpenPAKFile(var PAKFile: file; const PAKName: string): Boolean;
begin
  FileMode := fmOpenWrite;

  {$I-}
  AssignFile(PAKFile, PAKName);
  Rewrite(PAKFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function MakeValidName(var FileInfo: TFileInfo): Boolean;
var
  i: Longint;
begin
  Result := True;

  with FileInfo do
  begin
    FSName := AnsiLowerCase(FSName);

    for i := 1 to Length(FSName) do
    begin
      if FSName[i] = '\' then
        FSName[i] := '/';

      Result := Result and (FSName[i] in NameTableSet)
    end
  end
end;

procedure ScanDir(const Dir, RelDir: string; var FileInfos: TFileInfos);
var
  SearchRec: TSearchRec;
begin
  if FindFirst(Format('%s*.*', [Dir]), faAnyFile, SearchRec) = 0 then
    repeat
      with SearchRec do
      begin
        if (SearchRec.Attr and faDirectory) <> 0 then
        begin
          if (Name = '.') or (Name = '..') then
            Continue;

          // recursion for subdirectories
          ScanDir(Format('%s%s\', [Dir, Name]), Format('%s%s\', [RelDir, Name]),
            FileInfos)
        end
        else
        begin
          // add file if possible
          SetLength(FileInfos, Length(FileInfos) + 1);
          with FileInfos[High(FileInfos)] do
          begin
            if Length(Dir) + Length(Name) > 255 then
            begin
              WriteLn(Format('Warning: Cannot add "%s" (path too long)',
                [RealName]));
              SetLength(FileInfos, Length(FileInfos) - 1);
              Continue
            end;

            FSName := Format('%s%s', [RelDir, Name]);
            RealName := Format('%s%s', [Dir, Name]);
            Size := SearchRec.Size;

            if not MakeValidName(FileInfos[High(FileInfos)]) then
            begin
              WriteLn(Format('Warning: Cannot add "%s" (invalid characters)',
                [RealName]));
              SetLength(FileInfos, Length(FileInfos) - 1);
              Continue
            end
          end
        end
      end
    until FindNext(SearchRec) <> 0;

  FindClose(SearchRec)
end;

function LastEqChar(const s1, s2: string): Longint;
begin
  Result := 0;

  while (Result < Length(s1)) and (Result < Length(s2)) and
    (s1[Result + 1] = s2[Result + 1]) do
    Inc(Result)
end;

function AddName(var Names: TNames; var NameIdxs: TNameIdxs;
  const Name: string): Longint;
var
  i, j    : Longint;
  TempName: string;
begin
  for i := 0 to High(Names) do
  begin
    // get next name
    TempName := '';
    j := 0;
    while Names[i + j] <> 0 do
    begin
      TempName := TempName + NameTable[Names[i + j]];
      Inc(j)
    end;

    if TempName = Name then
    begin
      Result := i;
      Exit
    end
  end;

  // name not found
  Result := High(Names) + 1;
  SetLength(NameIdxs, Length(NameIdxs) + 1);
  NameIdxs[High(NameIdxs)] := Result;

  SetLength(Names, Length(Names) + Length(Name) + 1);
  for i := 0 to Length(Name) - 1 do
    Names[Result + i] := NameTableInv[Ord(Name[i + 1])];
  Names[High(Names)] := 0
end;

procedure PrepareBaseNode(const FileInfos: TFileInfos;
  FileInfoLeft, FileInfoRight: Longint; var HashNodes: THashNodes;
  var Allocated: TAllocated; var BaseNodeIdx: Longint);
var
  i, j,
  Al,
  HighAl,
  OldIdx  : Longint;
  SubAlloc: array[1..43] of Boolean;
begin
  // clear sub-allocation
  HighAl := 0;
  OldIdx := BaseNodeIdx;
  for i := 1 to 43 do
    SubAlloc[i] := False;

  // scan all characters-to-be-hashed
  for i := FileInfoLeft to FileInfoRight do
    if Length(FileInfos[i].FSName) > 0 then
    begin
      Al := NameTableInv[Ord(FileInfos[i].FSName[1])];

      SubAlloc[Al] := True;
      if Al > HighAl then
        HighAl := Al
    end;

  // search for possible base node position
  for i := 0 to High(HashNodes) do
  begin
    BaseNodeIdx := i;

    for j := 1 to HighAl do
    begin
      if not SubAlloc[j] then
        Continue;

      if (i + j <= High(HashNodes)) and Allocated[i + j] then
      begin
        BaseNodeIdx := -1;
        Break
      end
    end;

    if BaseNodeIdx <> -1 then
      Break
  end;

  // expand hash node table if necessary
  if BaseNodeIdx + HighAl > High(HashNodes) then
  begin
    Al := High(HashNodes);

    SetLength(HashNodes, BaseNodeIdx + HighAl + 1);
    SetLength(Allocated, Length(HashNodes));

    for i := Al + 1 to High(HashNodes) do
    begin
      with HashNodes[i] do
      begin
        Offset := 0;
        Size := 0;
        BaseNodeIdx := 0;
        PathLen := 0;
        NameIdx := 0
      end;

      Allocated[i] := False
    end
  end;

  // set base node and reserve hashes
  HashNodes[OldIdx].BaseNodeIdx := BaseNodeIdx;
  for i := 1 to 43 do
    if SubAlloc[i] then
      Allocated[BaseNodeIdx + i] := True
end;

procedure CreateHashTree(const FileInfos: TFileInfos;
  LeftIdx, RightIdx: Longint; var HashNodes: THashNodes;
  var Allocated: TAllocated; var Names: TNames; var NameIdxs: TNameIdxs;
  BaseNodeIdx, CurPathLen: Longint);
var
  i, j,
  LastHashPos,
  LastEqPos,
  NodeIdx     : Longint;
  LastHashChar: Char;
  CurName     : string;
begin
  // anything to process?
  if LeftIdx > RightIdx then
    Exit;

  // single entry is equivalent to file node
  if LeftIdx = RightIdx then
  begin
    WriteLn('Adding: ', FileInfos[LeftIdx].RealName);

    FileInfos[LeftIdx].NodeIdx := BaseNodeIdx;

    Inc(HashNodes[BaseNodeIdx].PathLen, 1);

    Exit
  end;

  // get good position for new base node
  PrepareBaseNode(FileInfos, LeftIdx, RightIdx, HashNodes, Allocated,
    BaseNodeIdx);

  // hash names
  i := LeftIdx;
  while i <= RightIdx do
  begin
    // file name substring matches are not allowed
    if Length(FileInfos[i].FSName) = 0 then
    begin
      WriteLn(Format('Warning: Cannot add "%s" (substring found)',
        [FileInfos[i].RealName]));
      FileInfos[i].NodeIdx := -1;

      Inc(i);
      Continue
    end;

    // hash by first character
    LastHashChar := FileInfos[i].FSName[1];
    LastHashPos := i;

    // get last match for current hash
    for j := i + 1 to RightIdx do
      if FileInfos[j].FSName[1] = LastHashChar then
        Inc(i)
      else
        Break;

    // start comparison at first name
    CurName := FileInfos[LastHashPos].FSName;

    // search for common prefix
    LastEqPos := Length(CurName);
    for j := LastHashPos + 1 to i do
    begin
      LastEqPos := LastEqChar(CurName, FileInfos[j].FSName);
      CurName := Copy(CurName, 1, LastEqPos)
    end;

    // shorten names by common prefix
    for j := LastHashPos to i do
      with FileInfos[j] do
        FSName := Copy(FSName, LastEqPos + 1, Length(FSName) - LastEqPos);

    // add prefix and set node infos
    NodeIdx := BaseNodeIdx + NameTableInv[Ord(LastHashChar)];
    with HashNodes[NodeIdx] do
    begin
      PathLen := CurPathLen + Length(CurName) - 1;
      NameIdx := AddName(Names, NameIdxs, Copy(CurName, 2, Length(CurName) - 1))
    end;

    // proceed recursively
    CreateHashTree(FileInfos, LastHashPos, i, HashNodes, Allocated, Names,
      NameIdxs, NodeIdx, CurPathLen + Length(CurName));

    // next file
    Inc(i)
  end
end;

procedure SetFileData(const FileInfos: TFileInfos; var HashNodes: THashNodes;
  Offset: Longint);
var
  i: Longint;
begin
  // set all node offsets to base value
  for i := 0 to High(HashNodes) do
    HashNodes[i].Offset := Offset;

  // compute file offsets
  for i := 0 to High(FileInfos) do
    if FileInfos[i].NodeIdx <> -1 then
    begin
      HashNodes[FileInfos[i].NodeIdx].Offset := Offset;
      HashNodes[FileInfos[i].NodeIdx].Size := FileInfos[i].Size;
      Inc(Offset, FileInfos[i].Size)
    end
end;

procedure CopyFiles(var PAKFile: file; const FileInfos: TFileInfos;
  const HashNodes: THashNodes);
var
  i     : Longint;
  InFile: file;
  Buffer: array of Byte;
begin
  for i := 0 to High(FileInfos) do
    if FileInfos[i].NodeIdx <> -1 then
    begin
      WriteLn(Format('Copying: %s', [FileInfos[i].RealName]));

      // allocate buffer
      SetLength(Buffer, FileInfos[i].Size);

      // open input file
      FileMode := fmOpenRead;
      {$I-}
      AssignFile(InFile, FileInfos[i].RealName);
      Reset(InFile, 1);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('  file open error');
        Continue
      end;

      // read file data to buffer
      {$I-}
      BlockRead(InFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('  file read error');
        CloseFile(InFile);
        Continue
      end;

      // go to computed offset
      {$I-}
      Seek(PAKFile, HashNodes[FileInfos[i].NodeIdx].Offset);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn(Format('  file seek error (looked for offset %d)',
          [HashNodes[FileInfos[i].NodeIdx].Offset]));
        CloseFile(InFile);
        Continue
      end;

      // write file data from buffer
      {$I-}
      BlockWrite(PAKFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('  PAK write error');
        CloseFile(InFile);
        Continue
      end;

      // clean up
      {$I-}
      CloseFile(InFile);
      {$I+}
      if IOResult <> 0 then
        WriteLn('  file close error')
    end
end;

function Less(var DataA, DataB): Boolean;
begin
  Less := (TFileInfo(DataA).FSName < TFileInfo(DataB).FSName)
end;

procedure Pad(var PAKFile: file);
const
  PadSize = 131072;
var
  Buffer: array[0..PadSize - 1] of Byte;
begin
  FillChar(Buffer[0], PadSize, $ae);

  {$I-}
  BlockWrite(PAKFile, Buffer[0],
    (PadSize - FileSize(PAKFile) mod PadSize) mod PadSize);
  {$I+}
  if IOResult <> 0 then
    WriteLn('  PAK write error')
end;

procedure CreatePAK(BaseDir, PAKName: string);
var
  FileInfos: TFileInfos;
  PAKHeader: TPAKHeader;
  HashNodes: THashNodes;
  Allocated: TAllocated;
  Names    : TNames;
  NameIdxs : TNameIdxs;
  BaseOffs : Longint;
  PAKFile  : file;
begin
  // validate requested base directory
  BaseDir := IncludeTrailingPathDelimiter(ExpandFileName(BaseDir));
  if not DirectoryExists(BaseDir) then
  begin
    WriteLn(Format('"%s" is not a valid directory', [BaseDir]));
    Exit
  end;

  // create PAK file
  if not OpenPAKFile(PAKFile, PAKName) then
  begin
    WriteLn(Format('Could not write to %s', [PAKName]));
    Exit
  end;

  // collect file data from base directory
  SetLength(FileInfos, 0);
  ScanDir(BaseDir, '', FileInfos);

  // sort file entries
  QuickSort(FileInfos, SizeOf(TFileInfo), Less);

  // prepare root hash node
  SetLength(HashNodes, 1);
  with HashNodes[0] do
  begin
    Offset := 0;
    Size := 0;
    PathLen := 0;
    BaseNodeIdx := 0;
    NameIdx := 0
  end;

  // root is allocated
  SetLength(Allocated, 1);
  Allocated[0] := True;

  // set up empty name table
  SetLength(Names, 0);
  SetLength(NameIdxs, 0);

  // recursively generate hash nodes
  CreateHashTree(FileInfos, 0, High(FileInfos), HashNodes, Allocated, Names,
    NameIdxs, 0, 0);

  // set header data
  with PAKHeader do
  begin
    ID := 'tlj_pack0001';
    NumHashNodes := Length(HashNodes);
    NumNameIdxs := Length(NameIdxs);
    NumNames := Length(Names)
  end;

  // align base offset at CD-ROM sector boundary
  BaseOffs := SizeOf(PAKHeader) + Length(HashNodes) * SizeOf(HashNodes[0]) +
    Length(Names) + Length(NameIdxs) * SizeOf(NameIdxs[0]);
  BaseOffs := (BaseOffs div 2048 + Byte(BaseOffs mod 2048 <> 0)) * 2048;

  // set file offset and size information for hash nodes
  SetFileData(FileInfos, HashNodes, BaseOffs);

  {$I-}
  BlockWrite(PAKFile, PAKHeader, SizeOf(PAKHeader));
  BlockWrite(PAKFile, HashNodes[0], Length(HashNodes) * SizeOf(HashNodes[0]));
  BlockWrite(PAKFile, Names[0], Length(Names));
  BlockWrite(PAKFile, NameIdxs[0], Length(NameIdxs) * SizeOf(NameIdxs[0]));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('Error: Could not write PAK structures.');
    CloseFile(PAKFile);
    Exit
  end;

  // insert file data into PAK
  CopyFiles(PAKFile, FileInfos, HashNodes);

  // fill up to even 128K boundary
  Pad(PAKFile);

  CloseFile(PAKFile)
end;

begin
  if ParamCount = 2 then
    CreatePAK(ParamStr(1), ParamStr(2))
  else
    WriteLn('Incorrect number of parameters.')
end.
